"""Main cloudwanderer module."""
import concurrent.futures
import logging
from typing import Callable, Iterator, List, NamedTuple

from .aws_interface import CloudWandererAWSInterface
from .cloud_wanderer_resource import CloudWandererResource
from .storage_connectors import BaseStorageConnector
from .urn import URN
from .utils import exception_logging_wrapper

logger = logging.getLogger("cloudwanderer")


class CloudWanderer:
    """CloudWanderer."""

    def __init__(
        self, storage_connectors: List["BaseStorageConnector"], cloud_interface: CloudWandererAWSInterface = None
    ) -> None:
        """Initialise CloudWanderer.

        Args:
            storage_connectors (List[BaseStorageConnector]):
                CloudWanderer storage connector objects.
            cloud_interface (CloudWandererAWSInterface):
                The cloud interface to get resources from.
                Defaults to :class:`~cloudwanderer.aws_interface.CloudWandererAWSInterface`.
        """
        self.storage_connectors = storage_connectors
        self.cloud_interface = cloud_interface or CloudWandererAWSInterface()

    def write_resource(self, urn: URN, **kwargs) -> None:
        """Fetch data for and persist to storage a single resource and its subresources.

        If the resource does not exist it will be deleted from the storage connectors.

        Arguments:
            urn (URN):
                The URN of the resource to write
            **kwargs:
                All additional keyword arguments will be passed down to the cloud interface client calls.
        """
        resources = list(self.cloud_interface.get_resource(urn=urn, **kwargs))

        for resource in resources:
            list(self._write_resource(resource=resource))
        if not resources:
            for storage_connector in self.storage_connectors:
                storage_connector.delete_resource(urn)

    def write_resources(
        self,
        regions: List[str] = None,
        service_names: List[str] = None,
        resource_types: List[str] = None,
        exclude_resources: List[str] = None,
        **kwargs,
    ) -> None:
        """Write all AWS resources in this account from all regions and all services to storage.

        All arguments are optional.

        Arguments:
            regions(list):
                The name of the region to get resources from (defaults to session default if not specified)
            service_names (str):
                The names of the services to write resources for (e.g. ``['ec2']``)
            resource_types (list):
                A list of resource types to include (e.g. ``['instance']``)
            exclude_resources (list):
                A list of service:resources to exclude (e.g. ``['ec2:instance']``)
            kwargs:
                All additional keyword arguments will be passed down to the cloud interface client calls.

        """
        urns = []
        actions = self.cloud_interface.get_actions(
            regions=regions,
            service_names=service_names,
            resource_types=resource_types,
            exclude_resources=exclude_resources,
        )
        for action_set in actions:
            for get_action in action_set.get_actions:
                resources = self.cloud_interface.get_resources(
                    region=get_action.region,
                    service_name=get_action.service_name,
                    resource_type=get_action.resource_type,
                )
                for resource in resources:
                    urns.extend(list(self._write_resource(resource)))
            for cleanup_action in action_set.cleanup_actions:
                for storage_connector in self.storage_connectors:
                    storage_connector.delete_resource_of_type_in_account_region(
                        account_id=self.cloud_interface.account_id,
                        region=cleanup_action.region,
                        service=cleanup_action.service_name,
                        resource_type=cleanup_action.resource_type,
                        urns_to_keep=urns,
                    )

    def write_resources_concurrently(
        self,
        cloud_interface_generator: Callable,
        storage_connector_generator: Callable,
        exclude_resources: List[str] = None,
        concurrency: int = 10,
        **kwargs,
    ) -> List["CloudWandererConcurrentWriteThreadResult"]:
        """Write all AWS resources in this account from all regions and all services to storage.

        Any additional args will be passed into the cloud interface's ``get_`` methods.
        **WARNING:** Experimental.

        Arguments:
            exclude_resources (list):
                exclude_resources (list): A list of service:resources to exclude (e.g. ``['ec2:instance']``)
            concurrency (int):
                Number of query threads to invoke concurrently.
            cloud_interface_generator (Callable):
                 A method which returns a new cloud interface session when called.
                This helps prevent non-threadsafe cloud interfaces from interfering with each others.
            storage_connector_generator (Callable):
                A method which returns a list of storage connectors when called.
                The returned connectors should be instances of the same connectors each time the method is called.
                These connectors do **not** need to be thread safe and will be returned at the end of execution.
            **kwargs:
                Additional keyword arguments will be passed down to the cloud interface methods.
        """
        logger.info("Writing resources in all regions")
        logger.warning("Using concurrency of: %s - CONCURRENCY IS EXPERIMENTAL", concurrency)
        with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
            threads = []
            for region_name in self.cloud_interface.enabled_regions:
                cw = CloudWanderer(
                    storage_connectors=storage_connector_generator(), cloud_interface=cloud_interface_generator()
                )
                threads.append(
                    executor.submit(
                        exception_logging_wrapper,
                        method=cw.write_resources,
                        exclude_resources=exclude_resources,
                        regions=[region_name],
                        return_value=cw.storage_connectors,
                        **kwargs,
                    )
                )
        thread_results: List[CloudWandererConcurrentWriteThreadResult] = []
        for thread in threads:
            result = thread.result()
            if result:
                thread_results.append(CloudWandererConcurrentWriteThreadResult(storage_connectors=result))
        return thread_results

    def _write_resource(self, resource: CloudWandererResource) -> Iterator[URN]:
        for storage_connector in self.storage_connectors:
            storage_connector.write_resource(resource)
        yield resource.urn


class CloudWandererConcurrentWriteThreadResult(NamedTuple):
    """The result from write_resources_concurrently."""

    storage_connectors: BaseStorageConnector



"""Standardised dataclasses for returning resources from storage."""
import logging
from typing import Callable, List, Optional

import jmespath  # type: ignore
from botocore import xform_name  # type: ignore

from .boto3_loaders import ServiceMap, ServiceMappingLoader
from .urn import URN
from .utils import snake_to_pascal

logger = logging.getLogger(__name__)


class ResourceMetadata:
    """Metadata for a :class:`CloudWandererResource`.

    Contains the original dictionaries of the resource and its attributes.

    Attributes:
        resource_data (dict):
            The raw dictionary representation of the Resource.
        secondary_attributes (list):
            The list of the resource's :class:`SecondaryAttribute` objects.
    """

    def __init__(self, resource_data: dict, secondary_attributes: List["SecondaryAttribute"]) -> None:
        """Initialise the data class.

        Arguments:
            resource_data (dict):
                The raw dictionary representation of the Resource.
            secondary_attributes (list):
                The list of the resource's :class:`SecondaryAttribute` objects.
        """
        self.resource_data = resource_data
        self.secondary_attributes = secondary_attributes


class CloudWandererResource:
    """A simple representation of a resource that prevents any storage metadata polluting the resource dictionary.

    This can be a resource, or a subresource.
    A subresource in CloudWanderer is a resource which does not have a unique identifier of its own and depends
    upon its parent for its identity.

    Attributes:
        urn: The URN of the resource.
        subresource_urns: The URNs of this resource's subresources (e.g. role_policies for a role).
        parent_urn: The URN of this resource's parent (only exists if this is a subresource).
        cloudwanderer_metadata (ResourceMetadata): The metadata of this resource (including attributes).
    """

    def __init__(
        self,
        urn: URN,
        resource_data: dict,
        secondary_attributes: List["SecondaryAttribute"] = None,
        loader: Optional[Callable] = None,
        subresource_urns: List[URN] = None,
    ) -> None:
        """Initialise the resource.

        Arguments:
            urn: The URN of the resource.
            subresource_urns: The URNs of the subresources of this resource.
            resource_data: The dictionary containing the raw data about this resource.
            secondary_attributes: A list of secondary attribute raw dictionaries.
            loader: The method which can be used to fulfil the :meth:`CloudWandererResource.load`.
        """
        self.urn = urn
        self.subresource_urns = subresource_urns or []
        self.cloudwanderer_metadata = ResourceMetadata(
            resource_data=resource_data or {}, secondary_attributes=secondary_attributes or []
        )
        service_mapping_loader = ServiceMappingLoader()
        self.cloudwanderer_service_metadata = ServiceMap.factory(
            name=urn.service,
            definition=service_mapping_loader.get_service_mapping(service_name=urn.service),
        )
        self.cloudwanderer_resource_metadata = self.cloudwanderer_service_metadata.get_resource_map(
            snake_to_pascal(urn.resource_type)
        )

        self._loader = loader
        self._set_resource_data_attrs()

    def load(self) -> None:
        """Inflate this resource with all data from the original storage connector it was spawned from.

        Raises:
            ValueError: Occurs if the storage connector loader isn't populated or the resource no longer exists
                in the StorageConnector's storage.
        """
        if self._loader is None:
            raise ValueError(f"Could not inflate {self}, storage connector loader not populated")
        updated_resource = self._loader(urn=self.urn)
        if updated_resource is None:
            raise ValueError(f"Could not inflate {self}, does not exist in storage")
            return
        self.cloudwanderer_metadata = updated_resource.cloudwanderer_metadata
        self._set_resource_data_attrs()

    @property
    def is_inflated(self) -> bool:
        """Return whether this resource has all the attributes from storage."""
        return bool([key for key in self.cloudwanderer_metadata.resource_data if not key.startswith("_")])

    @property
    def is_subresource(self) -> bool:
        """Return whether this resource is a subresource or not."""
        return self.cloudwanderer_resource_metadata.type == "subresource"

    @property
    def parent_urn(self) -> Optional[URN]:
        if not self.is_subresource:
            return None
        parent_resource_type = self.cloudwanderer_resource_metadata.parent_resource_type
        if not parent_resource_type:
            raise ValueError(f"{self.urn.resource_type} missing parent resource type in service map.")
        parent_resource_metadata = jmespath.search(
            f"resources.{snake_to_pascal(parent_resource_type)}", self.cloudwanderer_service_metadata.boto3_definition
        )
        parent_identifiers = parent_resource_metadata.get("identifiers")
        parent_slice = len(parent_identifiers)
        return URN(
            account_id=self.urn.account_id,
            region=self.urn.region,
            service=self.urn.service,
            resource_type=self.cloudwanderer_resource_metadata.parent_resource_type,
            resource_id_parts=self.urn.resource_id_parts[:parent_slice],
        )

    def get_secondary_attribute(self, name: str = None, jmes_path: str = None) -> List["SecondaryAttribute"]:
        """Get an attribute not returned in the resource's standard ``describe`` method.

        Arguments:
            name (str): The name of the secondary attribute (e.g. ``'enable_dns_support``)
            jmes_path (str): A JMES path to the secondary attribute. e.g. ``[].EnableDnsSupport.Value``
        """
        if name is not None:
            return [
                secondary_attr
                for secondary_attr in self.cloudwanderer_metadata.secondary_attributes
                if secondary_attr.name == name
            ]
        return jmespath.search(jmes_path, self.cloudwanderer_metadata.secondary_attributes)

    def _set_resource_data_attrs(self) -> None:
        for key, value in self.cloudwanderer_metadata.resource_data.items():
            if key.startswith("_"):
                continue
            setattr(self, xform_name(key), value)

    def __repr__(self) -> str:
        """Return a code representation of this resource."""
        return str(
            f"{self.__class__.__name__}("
            f"urn={repr(self.urn)}, "
            f"subresource_urns={repr(self.subresource_urns)}, "
            f"resource_data={self.cloudwanderer_metadata.resource_data}, "
            f"secondary_attributes={self.cloudwanderer_metadata.secondary_attributes})"
        )

    def __str__(self) -> str:
        """Return the string representation of this Resource."""
        return repr(self)


class SecondaryAttribute(dict):
    """Partially formalised SecondaryAttribute for resources.

    Allows us to store unstructured data in a dict-like object while maintaining the
    attribute_name attribute.

    Attributes:
        attribute_name (str): The name of the attribute.
    """

    def __init__(self, name: str, **kwargs) -> None:
        """Initialise the Secondary Attribute.

        Arguments:
            name (str): The name of the attribute
            **kwargs: The attributes keys and values
        """
        super().__init__(**kwargs)
        self.name = name



"""Boto3 Loaders.

Loaders and data classes to load custom CloudWanderer Boto3 service definitions and merge them with the boto3
provided ones. This allows cloudwanderer to extend Boto3 to support AWS resources that it does not support natively.
We can do this quite easily because CloudWanderer only needs a fraction of the functionality that native
Boto3 resources provide (i.e. the description of the resources).
"""
import logging
import os
import pathlib
from functools import lru_cache
from typing import Any, Dict, Generator, List, NamedTuple, Optional, Tuple

import boto3
import botocore  # type: ignore
import jmespath  # type: ignore
from boto3.resources.base import ServiceResource
from boto3.resources.model import Collection, ResourceModel
from botocore.exceptions import UnknownServiceError  # type: ignore

from .cache_helpers import cached_property, memoized_method
from .exceptions import UnsupportedServiceError
from .models import AWSGetAndCleanUp, CleanupAction, GetAction
from .utils import load_json_definitions

logger = logging.getLogger(__file__)


class CustomServiceLoader:
    """A class to load custom services."""

    def __init__(self, definition_path: str = "resource_definitions") -> None:
        self.service_definitions_path = os.path.join(pathlib.Path(__file__).parent.absolute(), definition_path)

    @cached_property
    def service_definitions(self) -> dict:
        """Return our custom resource definitions."""
        return load_json_definitions(self.service_definitions_path)

    def get_service_definition(self, service_name: str) -> dict:
        try:
            return self.service_definitions[service_name]
        except KeyError:
            raise UnsupportedServiceError(f"{service_name} does not exist as a custom CloudWanderer service.")

    @property
    def available_services(self) -> List[str]:
        """Return a list of available snake_case service names."""
        return list(self.service_definitions.keys())


class MergedServiceLoader:
    """A class to merge the services from a custom service loader with those of Boto3."""

    @lru_cache()  # type: ignore
    def __init__(self) -> None:
        self.custom_service_loader = CustomServiceLoader()
        botocore_session = botocore.session.get_session()
        self.boto3_loader = botocore_session.get_component("data_loader")
        self.boto3_loader.search_paths.append(os.path.join(os.path.dirname(boto3.__file__), "data"))

        # This does not need to honour the user's session because it is *only* used to list resources
        self.non_specific_boto3_session = boto3.session.Session()

    @property
    def available_services(self) -> List[str]:
        """Return a list of service names that can be loaded."""
        return list(set(self.cloudwanderer_available_services + self.boto3_available_services))

    @property
    def cloudwanderer_available_services(self) -> List[str]:
        """Return a list of services defined by CloudWanderer."""
        return self.custom_service_loader.available_services

    @property
    def boto3_available_services(self) -> List[str]:
        """Return a list of services defined by Boto3."""
        return self.non_specific_boto3_session.get_available_resources()

    @memoized_method()
    def get_service_definition(self, service_name: str) -> dict:
        """Return a combined dictionary service definition of both CloudWanderer and Boto3 services.

        Arguments:
            service_name: The PascalCase name of the service to return the definition for.

        Raises:
            UnsupportedServiceError: Occurs if a service is requested that CloudWanderer does not support.
        """
        boto3_definition = self._get_boto3_definition(service_name)
        custom_service_definition = self._get_custom_service_definition(service_name)

        if not boto3_definition and not custom_service_definition:
            raise UnsupportedServiceError(
                f"{service_name} is not supported by either Boto3 or CloudWanderer's custom services"
            )

        return {
            "service": {
                "hasMany": {
                    **(jmespath.search("service.hasMany", boto3_definition) or {}),
                    **(jmespath.search("service.hasMany", custom_service_definition) or {}),
                }
            },
            "resources": {**boto3_definition.get("resources", {}), **custom_service_definition.get("resources", {})},
        }

    def _get_custom_service_definition(self, service_name: str) -> dict:
        """Get the custom CloudWanderer definition for service_name so we can merge it with Boto3's.

        Arguments:
            service_name: The PascalCase name of the service to get the definition of.
        """
        try:
            return self.custom_service_loader.get_service_definition(service_name=service_name)
        except UnsupportedServiceError:
            return {}

    def _get_boto3_definition(self, service_name: str) -> dict:
        """Get the Boto3 definition for service_name so we can merge it with CloudWanderer's.

        Arguments:
            service_name (str): The name of the service (e.g. ``'ec2'``)
        """
        try:
            return self.boto3_loader.load_service_model(service_name, "resources-1", None)
        except UnknownServiceError:
            return {}


class ServiceMappingLoader:
    """A class to load and retrieve service mappings.

    Service Mappings provide additional metadata about an AWS service.
    This includes things like, whether it is a global service, whether it has regional resources, etc.
    """

    @lru_cache()  # type: ignore
    def __init__(self) -> None:
        """Load and retrieve service mappings."""
        self.service_mappings_path = os.path.join(pathlib.Path(__file__).parent.absolute(), "service_mappings")

    def get_service_mapping(self, service_name: str) -> dict:
        """Return the mapping for service_name.

        Arguments:
            service_name (str): The name of the service (e.g. ``'ec2'``)
        """
        return self.service_maps.get(service_name, {})

    @cached_property
    def service_maps(self) -> Dict[str, Any]:
        """Return our custom resource definitions."""
        return load_json_definitions(self.service_mappings_path)


class ServiceMap(NamedTuple):
    """Specification for additional CloudWanderer specific metadata about a Boto3 service."""

    name: str
    resource_definition: dict
    global_service: bool
    global_service_region: str
    service_definition: dict
    boto3_definition: dict

    @property
    def is_global_service(self) -> bool:
        """Return whether this service is a global service."""
        return self.global_service

    @property
    def is_default_service(self) -> bool:
        """Return True if this service has no definition and should return default values."""
        return self.service_definition == {}

    def get_resource_map(self, boto3_resource_name: str) -> "ResourceMap":
        """Return the resource map given a PascalCase resource name.

        Arguments:
            boto3_resource_name: The (PascalCase) name of the resource map to get.
        """
        return ResourceMap.factory(
            service_map=self,
            definition=self.resource_definition.get(boto3_resource_name, {}),
            boto3_resource_model=ResourceModel(
                name=boto3_resource_name,
                definition=self.boto3_definition["resources"].get(boto3_resource_name, {}),
                resource_defs=self.boto3_definition["resources"],
            ),
        )

    @classmethod
    def factory(cls, name: str, definition: dict) -> "ServiceMap":
        service_definition = definition.get("service", {})
        return cls(
            name=name,
            global_service=service_definition.get("globalService", False),
            global_service_region=service_definition.get("globalServiceRegion"),
            service_definition=service_definition,
            resource_definition=definition.get("resources", {}),
            boto3_definition=MergedServiceLoader().get_service_definition(service_name=name),
        )


class ResourceMap(NamedTuple):
    """Specification for additional CloudWanderer specific metadata about a Boto3 resource type.

    Attributes:
        region_request:
            An optional definition for how to perform a secondary query to
            discover the region in which this resource exists.
        parent_resource_type:
            The snake_case resource type of the parent (if this is a subresource).
        ignored_subresources:
            A list of snake case subresources which exist in the Boto3 definition but should be ignored.
        boto3_resource_model:
            The Boto3 model for this resource.
        default_filters:
            A dict of arguments to supply to the API Method used when enumerating this resource type.
            This can be overridden by the user with the ``filters`` argument.
        service_map:
            A link back to the parent :class:`ServiceMap` object.
        requires_load_for_full_metadata:
            If set ``resource.load()`` is called when fetching a list of these resources.
            Resources in a big increase in the number of API calls as one must be made for *each* resource.
        regional_resource:
            Whether or not this resource exists in every region.
    """

    type: Optional[str]
    region_request: Optional["ResourceRegionRequest"]
    resource_type: str
    parent_resource_type: str
    ignored_subresources: list
    boto3_resource_model: ResourceModel
    default_filters: Dict[str, Any]
    service_map: ServiceMap
    requires_load_for_full_metadata: bool = False
    regional_resource: bool = True

    @classmethod
    def factory(
        cls,
        service_map: ServiceMap,
        definition: Dict[str, Any],
        boto3_resource_model: ResourceModel,
    ) -> "ResourceMap":
        return cls(
            type=definition.get("type"),
            region_request=ResourceRegionRequest.factory(definition.get("regionRequest")),
            ignored_subresources=definition.get("ignoredSubresources", []),
            resource_type=botocore.xform_name(boto3_resource_model.name),
            parent_resource_type=definition.get("parentResourceType", ""),
            requires_load_for_full_metadata=definition.get("requiresLoadForFullMetadata", False),
            regional_resource=definition.get("regionalResource", True),
            default_filters=definition.get("defaultFilters", {}),
            service_map=service_map,
            boto3_resource_model=boto3_resource_model,
        )

    def should_query_resources_in_region(self, region: str) -> bool:
        """Return whether this resource should be queried from this region.

        Arguments:
            region: The region in which to query the resource.
        """
        if not self.service_map.is_global_service:
            return True
        return self.service_map.is_global_service and self.service_map.global_service_region == region

    def get_and_cleanup_actions(self, query_region: str) -> AWSGetAndCleanUp:
        """Return the query and cleanup actions to be performed if getting this resource type in this region.

        Arguments:
            query_region: The region in which the query would be performed.
        """
        actions = AWSGetAndCleanUp([], [])
        if not self.should_query_resources_in_region(query_region):
            return actions
        if self.service_map.is_global_service and self.regional_resource:
            cleanup_region = "ALL_REGIONS"
        else:
            cleanup_region = query_region
        if self.type != "subresource":
            actions.get_actions.append(
                GetAction(
                    service_name=self.service_map.name,
                    region=query_region,
                    resource_type=self.resource_type,
                )
            )
        actions.cleanup_actions.append(
            CleanupAction(
                service_name=self.service_map.name,
                region=cleanup_region,
                resource_type=self.resource_type,
            )
        )
        return actions

    @property
    def subresource_types(self) -> List[str]:
        """Return a list of CloudWanderer style subresource types it's possible for this resource type to have."""
        types = []
        for subresource_model in self.subresource_models:
            if not subresource_model.resource:
                continue
            types.append(botocore.xform_name(subresource_model.resource.model.name))
        return types

    @property
    def subresource_models(self) -> Generator[Collection, None, None]:
        """Yield the Boto3 models for the CloudWanderer style subresources of this resource type."""
        models = {}
        for collection_resource_map, collection_model in self._boto3_collection_models:
            response_resource = collection_model.resource
            if not response_resource or not self._is_collection_model_a_subresource(
                collection_resource_map, response_resource.model
            ):
                continue

            collection_resource_name = botocore.xform_name(response_resource.model.name)
            models[collection_resource_name] = collection_model
        yield from models.values()

    def _is_collection_model_a_subresource(
        self, collection_resource_map: "ResourceMap", resource_model: ResourceModel
    ) -> bool:

        if collection_resource_map.type in ["secondaryAttribute", "baseResource"] or resource_model is None:
            return False
        collection_resource_name = botocore.xform_name(resource_model.name)

        if len(resource_model.identifiers) == 1 and collection_resource_map.type != "subresource":
            return False
        if resource_model.name in self.ignored_subresource_types:
            logger.debug(
                "% is defined as an ignored subresource type by the %s servicemap",
                collection_resource_name,
                self.service_map.name,
            )
            return False

        return True

    @property
    def _boto3_collection_models(self) -> Generator[Tuple["ResourceMap", Collection], None, None]:
        """Yield the ResourceMaps and Collections for this resource type.

        This is used exclusively for subresources because subresources have a 1:many relationship with
        their parent resource.
        """
        for boto3_collection in self.boto3_resource_model.collections:
            if boto3_collection.resource is None:
                continue
            collection_resource_map = self.service_map.get_resource_map(boto3_collection.resource.model.name)
            yield collection_resource_map, boto3_collection

    @property
    def ignored_subresource_types(self) -> List:
        """Return a list of (PascalCase) ignored subresource types."""
        return [ignored_subresource["type"] for ignored_subresource in self.ignored_subresources]


class ResourceRegionRequest(NamedTuple):
    """Specification for a request to get a resource's region."""

    operation: str
    params: list
    path_to_region: str
    default_value: str

    @classmethod
    def factory(cls, definition: Optional[Dict[str, Any]]) -> Optional["ResourceRegionRequest"]:
        if not definition:
            return None
        return cls(
            operation=definition["operation"],
            params=[ResourceRegionRequestParam.factory(param) for param in definition["params"]],
            path_to_region=definition["pathToRegion"],
            default_value=definition["defaultValue"],
        )

    def build_params(self, resource: ServiceResource) -> dict:
        """Return the params required to query the resource's region from the Boto3 resource's attributes.

        Arguments:
            resource: The Boto3 Resource to build params for.
        """
        params = {}
        for param in self.params:
            params[param.target] = self._get_param_value(resource, param)
        return params

    def _get_param_value(self, resource: ServiceResource, param: "ResourceRegionRequestParam") -> Any:
        if param.source == "resourceAttribute":
            return getattr(resource, param.name)
        raise AttributeError(f"Invalid param source {param.source}")


class ResourceRegionRequestParam(NamedTuple):
    """Specification for a request param data model."""

    target: str
    source: str
    name: str

    @classmethod
    def factory(cls, definition: dict) -> "ResourceRegionRequestParam":
        return cls(target=definition["target"], source=definition["source"], name=definition["name"])



"""Helper classes and methods for interacting with boto3."""
import logging
from typing import List

import boto3
import botocore  # type: ignore

from .cache_helpers import cached_property

logger = logging.getLogger(__name__)


class Boto3CommonAttributesMixin:
    """Mixin that provides common informational attributes unique to boto3."""

    def __init__(self) -> None:
        self.boto3_session = boto3.session.Session()

    @cached_property
    def account_id(self) -> str:
        """Return the AWS Account ID our Boto3 session is authenticated against."""
        sts = self.boto3_session.client("sts")
        return sts.get_caller_identity()["Account"]

    @cached_property
    def region_name(self) -> str:
        """Return the default AWS region."""
        return self.boto3_session.region_name

    @cached_property  # type: ignore
    def enabled_regions(self) -> List[str]:
        """Return a list of enabled regions in this account."""
        regions = self.boto3_session.client("ec2").describe_regions()["Regions"]
        return [region["RegionName"] for region in regions if region["OptInStatus"] != "not-opted-in"]


def get_shape(boto3_resource: boto3.resources.base.ServiceResource) -> botocore.model.Shape:
    """Return the Botocore shape of a boto3 Resource.

    Parameters:
        boto3_resource (boto3.resources.base.ServiceResource):
            The resource to get the shape of.
    """
    service_model = boto3_resource.meta.client.meta.service_model
    shape = service_model.shape_for(boto3_resource.meta.resource_model.shape)
    return shape


def _clean_boto3_metadata(boto3_metadata: dict) -> dict:
    """Remove unwanted keys from boto3 metadata dictionaries.

    Arguments:
        boto3_metadata (dict): The raw dictionary of metadata typically found in resource.meta.data
    """
    boto3_metadata = boto3_metadata or {}
    unwanted_keys = ["ResponseMetadata"]
    for key in unwanted_keys:
        if key in boto3_metadata:
            del boto3_metadata[key]
    return boto3_metadata



"""A dataclass for building and querying AWS URNs.

AWS URNs are a standardised string format (name borrowed from Pulumi) which provide all the information required to
find a resource in AWS, whereas AWS ARNs do not always provide this information.

.. code-block ::

    # Format
    'urn:aws:<account_id>:<region>:<service>:<resource_type>:<resource_id>'
    # e.g.
    'urn:aws:111111111111:eu-west-2:iam:vpc:vpc-11111111'

Example:
    >>> from cloudwanderer import URN
    >>> URN.from_string('urn:aws:111111111111:eu-west-2:iam:vpc:vpc-11111111')
    URN(account_id='111111111111', region='eu-west-2', service='iam', \
resource_type='vpc', resource_id_parts=['vpc-11111111'])

"""
import re
from typing import Any, List, Optional, Union


class URN:
    """A dataclass for building and querying AWS URNs."""

    def __init__(
        self,
        account_id: str,
        region: str,
        service: str,
        resource_type: str,
        resource_id: Optional[str] = None,
        resource_id_parts: Optional[list] = None,
        cloud_name: str = None,
    ) -> None:
        """Initialise an AWS Urn.

        Arguments:
            account_id: AWS Account ID (e.g. ``111111111111``).
            region: AWS region (e.g. `us-east-1``).
            service: AWS Service (e.g. ``iam``).
            resource_type: AWS Resource Type (e.g. ``role_policy``)
            resource_id: AWS Resource Id (e.g. ``test-role-policy``)
            resource_id_parts: AWS Resource Id (e.g. ``test-role-policy``)
            cloud_name: The name of the cloud this resource exists in (defaults to ``'aws'``)

        Attributes:
            account_id: AWS Account ID (e.g. ``111111111111``).
            region: AWS region (e.g. `us-east-1``).
            service: AWS Service (e.g. ``iam``).
            resource_type: AWS Resource Type (e.g. ``role_policy``)
            resource_id: AWS Resource Id (e.g. ``test-role-policy``)
            resource_id_parts: AWS Resource ID parts (e.g. ``['test-role', 'test-role-policy']``)
            cloud_name: The name of the cloud this resource exists in (defaults to ``'aws'``)

        Raises:
            ValueError: If incorrect values are supplied.
        """
        self.cloud_name = cloud_name or "aws"
        self.account_id = account_id
        self.region = region
        self.service = service
        self.resource_type = resource_type
        resource_id_parts = [resource_id] if resource_id else resource_id_parts
        if not resource_id_parts or not all(resource_id_parts):
            raise ValueError("resource_id or id_parts must be supplied with non empty values")
        self.resource_id_parts: List[str] = resource_id_parts
        self.resource_id = "/".join(self.escape_id(id_part) or "" for id_part in self.resource_id_parts)

    @classmethod
    def from_string(cls, urn_string: str) -> "URN":
        """Create an URN Object from an URN string.

        Arguments:
            urn_string (str): The string version of an AWSUrn to convert into an object.

        Returns:
            URN: The instantiated AWS URN.

        Raises:
            ValueError: When no valid resource id found
        """
        parts = urn_string.split(":")
        try:
            resource_id_parts = re.split(r"(?<!\\)/", parts[6])
        except IndexError:
            raise ValueError("Resource ID must be supplied as the 7th element in a colon separated string")
        return cls(
            account_id=parts[2],
            region=parts[3],
            service=parts[4],
            resource_type=parts[5],
            resource_id_parts=[cls.unescape_id(id_part) for id_part in resource_id_parts],
        )

    @property
    def resource_id_parts_parsed(self) -> List[Union[str, int]]:
        """Return the URNs ID parts parsed to their original types where possible."""
        return [int(part) if part.isnumeric() else part for part in self.resource_id_parts]

    @staticmethod
    def unescape_id(escaped_id: Optional[str]) -> Optional[str]:
        """Return an unescaped ID with the forward slashes unescaped.

        Arguments:
            escaped_id: The id to unescape.
        """
        if escaped_id is None:
            return None
        return escaped_id.replace(r"\/", r"/")

    @staticmethod
    def escape_id(unescaped_id: Optional[str]) -> Optional[str]:
        """Return an escaped ID with the forward slashes escaped.

        Arguments:
            unescaped_id: The id to escape.
        """
        if unescaped_id is None:
            return None
        return re.sub(r"(?<!\\)/", r"\/", unescaped_id)

    def __eq__(self, other: Any) -> bool:
        """Allow comparison of one URN to another.

        Arguments:
            other (Any): The other object to compare this one with.
        """
        return str(self) == str(other)

    def __repr__(self) -> str:
        """Return a class representation of the URN."""
        return str(
            f"{self.__class__.__name__}("
            f"account_id='{self.account_id}', "
            f"region='{self.region}', "
            f"service='{self.service}', "
            f"resource_type='{self.resource_type}', "
            f"resource_id_parts={self.resource_id_parts})"
        )

    def __str__(self) -> str:
        """Return a string representation of the URN."""
        base = ":".join(["urn", self.cloud_name, self.account_id, self.region, self.service, self.resource_type])
        return f"{base}:{self.resource_id}"



"""Models for CloudWanderer data."""
from typing import Any, Dict, List, NamedTuple


class GetAction(NamedTuple):
    """A get action for a specific resource_type in a specific region.

    Attributes:
        service_name: The name of the service to get
        resource_type: The type of resource to get
        region: The region to get it in
    """

    service_name: str
    region: str
    resource_type: str


class CleanupAction(NamedTuple):
    """A storage connector clean up action for a specific resource_type in a specific region.

    Attributes:
        service_name: The name of the service to cleanup
        resource_type: The type of resource to cleanup
        region: The region to cleanup from storage
    """

    service_name: str
    region: str
    resource_type: str


class GetAndCleanUp(NamedTuple):
    """A set of get and cleanup actions.

    The get and cleanup actions are paired together because the cleanup action must be performed after the get action
    in order to ensure that any stale resources are removed from the storage connectors.

    It's possible for there to be far more cleanup actions than get actions if a single region API
    discovers resources in multiple regions.

    Attributes:
        get_actions: The list of get actions
        cleanup_actions: The list of cleanup actions
    """

    get_actions: List[GetAction]
    cleanup_actions: List[CleanupAction]

    def __add__(self, other: Any) -> "GetAndCleanUp":
        """Combine the actions of two objects.

        Arguments:
            other: the other GetAndCleanUp objects whose actions we will add to this one.

        Raises:
            TypeError: If anything other than a GetAndCleanUp action is added.
        """
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"unsupported operand type(s) for +: {self.__class__.__name__} and {other.__class__.__name__}"
            )
        self.get_actions.extend(other.get_actions)
        self.cleanup_actions.extend(other.cleanup_actions)
        return self


class AWSGetAndCleanUp(GetAndCleanUp):
    """An AWS specific set of GetAndCleanUp actions.

    This differs from a regular GetAndCleanUp action insofar as it
    will probably contain actions with the region 'ALL_REGIONS'.
    These actions need to be unpacked into region specific actions that
    reflect the enabled regions in the AWS account in question
    before being placed into a non-cloud specific GetAndCleanUp class
    for CloudWanderer to consume.
    """

    def inflate_actions(self, regions: List[str]) -> GetAndCleanUp:
        """Inflate the get and cleanup actions that are ALL_REGIONS.

        Arguments:
            regions: The list of enabled regions with which to inflate actions specified as ALL_REGIONS.
        """
        return GetAndCleanUp(
            get_actions=self._inflated_get_actions(regions), cleanup_actions=self._inflated_cleanup_actions(regions)
        )

    def _inflated_get_actions(self, regions: List[str]) -> List[GetAction]:
        get_actions = []
        for get_action in self.get_actions:
            if get_action.region != "ALL_REGIONS":
                get_actions.append(get_action)
                continue
            get_actions.extend(
                [
                    GetAction(
                        service_name=get_action.service_name, resource_type=get_action.resource_type, region=region
                    )
                    for region in regions
                ]
            )
        return get_actions

    def _inflated_cleanup_actions(self, regions: List[str]) -> List[CleanupAction]:
        cleanup_actions: List[CleanupAction] = []
        for cleanup_action in self.cleanup_actions:
            if cleanup_action.region != "ALL_REGIONS":
                cleanup_actions.append(cleanup_action)
                continue
            cleanup_actions.extend(
                [
                    CleanupAction(
                        service_name=cleanup_action.service_name,
                        resource_type=cleanup_action.resource_type,
                        region=region,
                    )
                    for region in regions
                ]
            )
        return cleanup_actions

    def __bool__(self) -> bool:
        """Return whether this GetAndCleanUp set is empty."""
        return bool(self.get_actions or self.cleanup_actions)


class ResourceFilter(NamedTuple):
    """A definition for filtering a resource."""

    service_name: str
    resource_type: str
    filters: Dict[str, Any]



"""Classes wrapping :class:`~boto3.resources.base.ServiceResource` objects.

The methods and arguments on these classes somtimes differ in name from those in Boto3's
Resources to make them easier to understand in this context..

Glossary:
    service_name:
        The snake_case name of an AWS service (e.g. ``ec2``)
    resource_type:
        The (unpluralised) snake_case name of the type of the resource in snake case (e.g. ``instance``)
    secondary_attribute:
        An attribute of a resource which requires a secondary API call to retrieve (e.g. VPC Attributes)
    subresource:
        A child resource which does not have its own ARN and must be queried by referencing
        its parent resource's ID (e.g. inline role policies). Different from Boto3 subresources which simply
        indicate a hierarchical relationship between resources (e.g. subnets are a child resource of vpcs).
"""

import logging
from typing import Any, Dict, Generator, List, NamedTuple, Optional, Tuple, Type

import boto3
import botocore  # type: ignore
import jmespath  # type: ignore
from boto3.resources.base import ServiceResource
from boto3.resources.factory import ResourceFactory
from boto3.resources.model import Action, Collection
from boto3.utils import ServiceContext

from .boto3_helpers import _clean_boto3_metadata, get_shape
from .boto3_loaders import MergedServiceLoader, ResourceMap, ServiceMap, ServiceMappingLoader
from .cache_helpers import cached_property, memoized_method
from .cloud_wanderer_resource import SecondaryAttribute
from .exceptions import (
    BadRequestError,
    BadServiceMapError,
    BadUrnAccountIdError,
    BadUrnIdentifiersError,
    BadUrnRegionError,
    ResourceNotFoundError,
    UnsupportedResourceTypeError,
)
from .urn import URN

logger = logging.getLogger(__name__)


class CloudWandererBoto3ResourceFactory:
    """Factory class for generating Boto3 Resource objects."""

    def __init__(self, boto3_session: boto3.session.Session = None) -> None:
        """Initialise the ResourceFactory.

        Arguments:
            boto3_session (boto3.session.Session): The :class:`boto3.session.Session` object to use for any queries.
        """
        self.boto3_session = boto3_session or boto3.session.Session()
        self.emitter = self.boto3_session.events
        self.factory = ResourceFactory(self.emitter)

    def load(self, service_name: str, resource_definitions: dict, service_definition: dict) -> Type:
        """Load the specified resource definition dictionaries into a Resource object.

        Arguments:
            service_name (str):
                The name of the service to load (e.g. ``'ec2'``)
            resource_definitions (dict):
                A dict describing the resource definitions.
                This is the ``'resources'`` key in each ``resource_definition`` json.
            service_definition (dict):
                A dict describing the service definition.
                This is the ``'service'`` key in each ``resource_definition`` json.
        """
        service_context = ServiceContext(
            service_name=service_name,
            resource_json_definitions=resource_definitions,
            service_model=self._get_service_model(service_name),
            service_waiter_model=None,
        )

        return self.factory.load_from_definition(
            resource_name=service_name,
            single_resource_json_definition=service_definition,
            service_context=service_context,
        )

    @memoized_method()
    def _get_service_model(self, service_name: str) -> botocore.model.ServiceModel:
        """Return the botocore service model corresponding to this service.

        Arguments:
            service_name: The service name to get the service model of.
        """
        logger.debug("Getting service model for %s", service_name)
        client = self.boto3_session.client(service_name=service_name)  # type: ignore
        return client.meta.service_model


class Boto3Services:
    """Wraps Boto3 Session.

    Allows us to:

    1. Wrap Boto3 ServiceResource objects with CloudWandererBoto3Service objects.
    2. Inject custom service definitions into our :class:`CloudWandererBoto3ResourceFactory` \
        and return them alongside the default Boto3 ServiceResource objects.

    Used by :class:`~cloudwanderer.aws_interface.CloudWandererAWSInterface` to instantiate
    services and can be used to get resources from their URN.
    """

    def __init__(
        self,
        boto3_session: boto3.session.Session = None,
        service_loader: MergedServiceLoader = None,
        service_mapping_loader: ServiceMappingLoader = None,
        account_id: str = None,
    ) -> None:
        """Initialise the Boto3SessionWrapper.

        Arguments:
            boto3_session:
                The :class:`boto3.session.Session` object to use for any queries.
            service_loader:
                Optionally specify your own service loader if you wish to insert your own resources.
            service_mapping_loader:
                Optionally specify your own service mapping loader if you wish to insert your own service mappings.
            account_id:
                Optionally specify your account id to save a call to STS.
        """
        self.boto3_session = boto3_session or boto3.session.Session()
        self._factory = CloudWandererBoto3ResourceFactory(boto3_session=self.boto3_session)
        self._loader = service_loader or MergedServiceLoader()
        self._service_mapping_loader = service_mapping_loader or ServiceMappingLoader()
        self._account_id = account_id

    @property
    def available_services(self) -> List[str]:
        """Return a list of service names that can be loaded by :meth:`Boto3Services.get_service`."""
        return self._loader.available_services

    @cached_property
    def account_id(self) -> str:
        """Return the AWS Account ID our Boto3 session is authenticated against."""
        if self._account_id:
            return self._account_id
        sts = self.boto3_session.client("sts")
        return sts.get_caller_identity()["Account"]

    def get_service(self, service_name: str, region_name: str = None, **kwargs) -> "CloudWandererBoto3Service":
        """Return the :class`CloudWandererBoto3Service` by this name.

        Arguments:
            service_name: The name of the service to instantiate.
            region_name: The region to instantiate the service for.
            **kwargs: Additional keyword args will be passed to the Boto3 client.
        """
        service_method = self._get_service_method(service_name)
        return CloudWandererBoto3Service(
            boto3_service=service_method(client=self._get_client(service_name, region_name=region_name, **kwargs)),
            service_map=ServiceMap.factory(
                name=service_name,
                definition=self._service_mapping_loader.get_service_mapping(service_name=service_name),
            ),
            account_id=self.account_id,
            enabled_regions=self.enabled_regions,
            region_name=region_name,
            boto3_session=self.boto3_session,
        )

    def get_empty_service(self, service_name: str, region_name: str = None) -> "CloudWandererBoto3Service":
        """Return the :class`CloudWandererBoto3Service` by this name without a Boto3 Client instantiated.

        Useful for querying service/resource metadata.

        Arguments:
            service_name: The name of the service to instantiate.
            region_name: The region to instantiate the service for.
        """
        logger.debug("Getting empty service for %s in %s", service_name, region_name)
        service_method = self._get_service_method(service_name)
        return CloudWandererBoto3Service(
            boto3_service=service_method(client=self._get_default_client(service_name)),
            service_map=self._get_service_map(service_name),
            account_id=self.account_id,
            enabled_regions=self.enabled_regions,
            region_name=region_name,
            boto3_session=self.boto3_session,
        )

    @memoized_method()
    def _get_service_map(self, service_name: str) -> ServiceMap:
        logger.debug("Getting service map for %s", service_name)
        return ServiceMap.factory(
            name=service_name,
            definition=self._service_mapping_loader.get_service_mapping(service_name=service_name),
        )

    @memoized_method()
    def _get_default_client(self, service_name: str) -> botocore.client.BaseClient:
        logger.debug("Getting default client for %s", service_name)
        return self._get_client(service_name=service_name, region_name="us-east-1")

    def _get_client(self, service_name: str, region_name: str = None, **kwargs) -> botocore.client.BaseClient:
        return self.boto3_session.client(service_name, region_name=region_name, **kwargs)  # type: ignore

    @memoized_method()
    def _get_service_method(self, service_name: str) -> Type:
        logger.debug("Getting service_method for %s", service_name)
        service_definition = self._loader.get_service_definition(service_name=service_name)
        return self._factory.load(
            service_name=service_name,
            service_definition=service_definition["service"],
            resource_definitions=service_definition["resources"],
        )

    def get_resource_from_urn(self, urn: URN) -> "CloudWandererBoto3Resource":
        """Return the :class:`CloudWandererBoto3Resource` resource picked out by this urn.

        Arguments:
            urn (URN): The urn of the resource to get.

        Raises:
            BadUrnAccountIdError: When the account ID of the URN does not match the account id of the current session.
        """
        if urn.account_id != self.account_id:
            raise BadUrnAccountIdError(f"{urn} exists in an account other than the current one ({self.account_id}).")

        service = self.get_service(urn.service, urn.region)
        return service.get_resource_from_urn(urn)

    @cached_property
    def enabled_regions(self) -> List[str]:
        """Return a list of enabled regions in this account."""
        regions = self.boto3_session.client("ec2").describe_regions()["Regions"]
        return [region["RegionName"] for region in regions if region["OptInStatus"] != "not-opted-in"]


class CloudWandererBoto3Service:
    """Wraps Boto3 :class:`~boto3.resources.base.ServiceResource` service-level objects.

    Allows us to include additional CloudWanderer specific functionality.
    The object represents an AWS service (e.g. ``ec2``) in a specific region.
    Used to get resources from the API as well as get metadata about the resource type from Boto3.
    """

    def __init__(
        self,
        boto3_service: ServiceResource,
        service_map: ServiceMap,
        account_id: str,
        boto3_session: boto3.session.Session,
        region_name: str = None,
        enabled_regions: List[str] = None,
    ) -> None:
        """Instantiate CloudWandererBoto3Service.

        Arguments:
            boto3_service: The boto3 service object to wrap.
            service_map: The CloudWanderer service map that provides additional context about this service.
            account_id: The ID of the AWS account our session is in.
            region_name: The region to get resources from for this service.
            boto3_session: The Boto3 session that created this client.
            enabled_regions: The list of regions currently enabled.
        """
        self.boto3_service = boto3_service
        self.boto3_session = boto3_session
        self.service_map = service_map
        self.account_id = account_id
        self.region_name = region_name
        self._enabled_regions = enabled_regions

    @property
    def resource_types(self) -> List[str]:
        """Return a list of snake_case resource types available on this service."""
        collection_resource_types = [
            collection.resource.type for collection in self._collections if collection.resource
        ]
        return [
            botocore.xform_name(resource.name)
            for resource in self._subresources
            if resource.resource and resource.resource.type in collection_resource_types
        ]

    @property
    def resource_summary(self) -> List["ResourceSummary"]:
        """Return a summary of resource types in this service."""
        summaries = []

        for resource_type in self.resource_types:
            resource = self._get_empty_resource(resource_type)
            if not resource:
                logger.debug("No %s resource type found %s", resource_type, self.name)
                continue
            summaries.append(
                ResourceSummary(
                    resource_type=resource_type,
                    resource_type_pascal=resource.resource_type_pascal,
                    service_friendly_name=self.friendly_name,
                    secondary_attribute_names=resource.secondary_attribute_names,
                    subresources=resource.subresource_summary,
                )
            )
        return summaries

    def _get_empty_resource(self, resource_type: str) -> Optional["CloudWandererBoto3Resource"]:
        """Return a resource object of resource_type which is not associated with a specific AWS resource.

        Useful for interrogating metadata about that type of resource.

        Arguments:
            resource_type: The CloudWanderer style (snake_case) resource type.
        """
        boto3_resource = self._get_boto3_resource(resource_type)
        boto3_resource_getter = getattr(self.boto3_service, boto3_resource.name)
        if not boto3_resource.resource:
            return None
        blank_args = ["" for identifier in boto3_resource.resource.identifiers]
        return CloudWandererBoto3Resource(
            account_id=self.account_id,
            cloudwanderer_boto3_service=self,
            boto3_resource=boto3_resource_getter(*blank_args),
            service_map=self.service_map,
        )

    def get_resource_from_urn(self, urn: URN) -> "CloudWandererBoto3Resource":
        """Return the :class:`CloudWandererBoto3Resource` resource picked out by this urn.

        Arguments:
            urn (URN): The urn of the resource to get.

        Raises:
            BadRequestError: Occurs when the AWS API returns a 4xx HTTP error other than 404.
            ResourceNotFoundError: Occurs when the AWS API Returns a 404 HTTP error.
            UnsupportedResourceTypeError: Occurs when the definition for the resource does not support loading by id.
            botocore.exceptions.ClientError: Boto3 Client Error
            BadUrnIdentifiersError: Occurs when the URN contains fewer identifiers than is required by the reource type.
            BadUrnRegionError: When the region of the URN is not possible with the service and/or resource type.
        """
        boto3_service_resource = self._get_boto3_resource(urn.resource_type)
        if not boto3_service_resource or not boto3_service_resource.resource:
            raise UnsupportedResourceTypeError(f"Resource type {urn.resource_type} not found")

        resource_map = self.service_map.get_resource_map(boto3_service_resource.name)
        if self.service_map.global_service_region != urn.region and not resource_map.regional_resource:
            raise BadUrnRegionError(f"{urn}'s service does not have resources in {urn.region}")

        boto3_resource_getter = getattr(self.boto3_service, boto3_service_resource.name)
        if len(urn.resource_id_parts) != len(boto3_service_resource.resource.identifiers):
            raise BadUrnIdentifiersError(
                f"An URN of type {urn.service} {urn.resource_type} should have "
                f"{len(boto3_service_resource.resource.identifiers)} "
                f"ID components instead of {len(urn.resource_id_parts)}"
            )
        boto3_resource = boto3_resource_getter(*urn.resource_id_parts_parsed)
        if not hasattr(boto3_resource, "load"):
            raise UnsupportedResourceTypeError(f"{urn.resource_type} does not support loading by ID.")

        try:
            boto3_resource.load()
        except botocore.exceptions.ClientError as ex:
            error_code = ex.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
            if error_code == 404:
                raise ResourceNotFoundError(f"{urn} was not found") from ex
            if error_code >= 400 and error_code < 500:
                raise BadRequestError(f"A request error was returned while fetching {urn}") from ex
            raise
        return CloudWandererBoto3Resource(
            account_id=self.account_id,
            cloudwanderer_boto3_service=self,
            boto3_resource=boto3_resource,
            service_map=self.service_map,
            resource_map=resource_map,
        )

    def _get_boto3_resource(self, resource_type: str) -> Action:
        """Return a callable Boto3 service resource of resource_type.

        Arguments:
            resource_type: The *snake_case singluar) resource type (e.g. ``instance``)
        """
        action = next(
            resource for resource in self._subresources if botocore.xform_name(resource.name) == resource_type
        )
        return action

    def get_resources(
        self, resource_type: str, resource_filters: Dict[str, Any] = None
    ) -> Generator["CloudWandererBoto3Resource", None, None]:
        """Yield all resources of resource_type in this region.

        Arguments:
            resource_type: The snake_case resource type to get.
            resource_filters: A resource filter object to apply when fetching resources.

        Raises:
            BadServiceMapError: Occurs if there is an error finding the collection for this resource_type.
        """
        collection_class = self._get_collection_from_resource_type(resource_type)
        if not collection_class or not collection_class.resource:
            raise BadServiceMapError(
                f"resource type '{resource_type}' has no Collection in service '{self.service_map.name}'. "
                "This may mean it's a subresource that hasn't been tagged as such in the resource map."
            )
        collection = getattr(self.boto3_service, collection_class.name)
        resource_map = self.service_map.get_resource_map(collection_class.resource.model.name)
        filters = resource_filters or resource_map.default_filters
        if filters:
            logger.info("Applying filter %s to resource %s", filters, resource_type)
            result = collection.filter(**filters)
        else:
            result = collection.all()
        for boto3_resource in result:
            if hasattr(boto3_resource, "load") and (
                not boto3_resource.meta.data or resource_map.requires_load_for_full_metadata
            ):
                boto3_resource.load()
            yield CloudWandererBoto3Resource(
                account_id=self.account_id,
                cloudwanderer_boto3_service=self,
                boto3_resource=boto3_resource,
                service_map=self.service_map,
                resource_map=resource_map,
            )

    def _get_collection_from_resource_type(self, resource_type: str) -> Optional[Collection]:
        """Return a collection given a CloudWanderer style (snake_case singular) resource type.

        The resource type we're passed in is NOT a Boto3 style resource type.
        It is actually a Boto3 subresource name that we need to first get the type of.

        Arguments:
            resource_type: The CloudWanderer style (snake_case) resource type name.
        """
        boto3_resource_type = self._get_boto3_resource_type_from_subresource_name(subresource_name=resource_type)
        if not boto3_resource_type:
            return None
        return self._get_collection_from_boto3_resource_type(boto3_resource_type)

    def _get_boto3_resource_type_from_subresource_name(self, subresource_name: str) -> Optional[str]:
        """Return the resource_type for a given subresource name.

        Resource names *almost* always match their resource type, but not always.
        I'm kooking at you EC2 KeyPair!

        Arguments:
            subresource_name: The snake_case CloudWanderer style subresource name.
        """
        return next(
            (
                resource.resource.type
                for resource in self._subresources
                if botocore.xform_name(resource.name) == subresource_name and resource.resource
            ),
            None,
        )

    def _get_collection_from_boto3_resource_type(self, boto3_resource_type: str) -> Optional[Collection]:
        """Return the Collection corresponding to a given Boto3 style (PascalCase) resource type.

        Arguments:
            boto3_resource_type: The PascalCase name for the Boto3 resource type.
        """
        return next(
            (
                collection
                for collection in self._collections
                if collection.resource and collection.resource.type == boto3_resource_type
            ),
            None,
        )

    @property
    def _collections(self) -> List[Collection]:
        return self.boto3_service.meta.resource_model.collections

    @property
    def _subresources(self) -> List[Action]:
        return self.boto3_service.meta.resource_model.subresources  # type: ignore

    @property
    def region(self) -> str:
        """Return the name of the region this service is querying."""
        if self.region_name is not None:
            return self.region_name
        client_region = self.boto3_service.meta.client.meta.region_name
        if client_region == "aws-global":
            return self.service_map.global_service_region or client_region
        return client_region

    @property
    def enabled_regions(self) -> List[str]:
        """Return a list of enabled regions in this account."""
        if not self._enabled_regions:
            regions = self.boto3_session.client("ec2").describe_regions()["Regions"]
            self._enabled_regions = [
                region["RegionName"] for region in regions if region["OptInStatus"] != "not-opted-in"
            ]
        return self._enabled_regions

    @property
    def name(self) -> str:
        """Return the snake_case name of this service."""
        return self.boto3_service.meta.service_name

    @property
    def friendly_name(self) -> str:
        """Return the friendly name of this service."""
        return self.boto3_service.meta.client.meta.service_model.metadata["serviceId"]


class CloudWandererBoto3Resource:
    """Wraps Boto3 :class:`~boto3.resources.base.ServiceResource` resource-level objects.

    Allows us to provide additional functionality specific to CloudWanderer.

    This is almost always tied to a specific resource that exists in AWS, but is occasionally instantiated
    abstractly in order to interrogate metadata about this resource type.
    """

    def __init__(
        self,
        account_id: str,
        boto3_resource: ServiceResource,
        cloudwanderer_boto3_service: CloudWandererBoto3Service,
        service_map: ServiceMap,
        resource_map: ResourceMap = None,
    ) -> None:
        self.account_id = account_id
        self.boto3_resource = boto3_resource
        self.cloudwanderer_boto3_service = cloudwanderer_boto3_service
        self.service_map = service_map
        self.resource_map = resource_map or self.service_map.get_resource_map(
            self.boto3_resource.meta.resource_model.name
        )

    @property
    def resource_type_pascal(self) -> str:
        """Return a friendly name for this resource type."""
        return self.boto3_resource.meta.resource_model.name

    @property
    def resource_type(self) -> str:
        """Return the snake_case resource type of this resource."""
        return botocore.xform_name(self.boto3_resource.meta.resource_model.name)

    @property
    def service(self) -> str:
        """Return the snake_case service type for this resource."""
        return self.service_map.name

    @property
    def id(self) -> str:
        """Return the resource ID."""
        return self.urn.resource_id

    @property
    def id_parts(self) -> List[str]:
        id_members = [x.name for x in self.boto3_resource.meta.resource_model.identifiers]
        raw_id_parts = [str(getattr(self.boto3_resource, id_member)) for id_member in id_members]
        id_parts = []
        for identifier in raw_id_parts:
            if identifier.startswith("arn:"):
                id_parts.append("".join(identifier.split(":")[5:]))
            else:
                id_parts.append(identifier)
        return id_parts

    @property
    def raw_data(self) -> dict:
        """Return the raw dictionary data for this resource."""
        return self.boto3_resource.meta.data

    @property
    def normalised_raw_data(self) -> dict:
        """Return the raw data ditionary for this resource, ensuring that all keys for this resource are present."""
        result = {attribute: None for attribute in get_shape(self.boto3_resource).members.keys()}
        result.update(self.raw_data or {})
        return _clean_boto3_metadata(result)

    @property
    def urn(self) -> URN:
        """Return the resource's Universal Resource Name."""
        return URN(
            account_id=self.account_id,
            region=self.region,
            service=self.service,
            resource_type=self.resource_type,
            resource_id_parts=self.id_parts,
        )

    @property
    def region(self) -> str:
        """Return the region for this resource.

        Typically this just takes the region of the session the resource was discovered from.
        However for some resources (e.g. S3 buckets) it performs an API call to look it up.
        """
        if not self.service_map.is_global_service:
            return self._boto3_client.meta.region_name
        if not self.resource_map.regional_resource:
            return self.service_map.global_service_region

        return self._get_region()

    @property
    def secondary_attribute_names(self) -> List[str]:
        """Return a list fof secondary attributes its possible for this resource type to have."""
        names = []
        for secondary_attribute_model in self.secondary_attribute_models:
            names.append(botocore.xform_name(secondary_attribute_model.name))
        return names

    def get_secondary_attributes(self) -> Generator[SecondaryAttribute, None, None]:
        """Return the secondary attributes for this resource."""
        for secondary_attribute_model in self.secondary_attribute_models:
            secondary_attribute = getattr(self.boto3_resource, secondary_attribute_model.name)()
            secondary_attribute.load()
            yield SecondaryAttribute(
                name=botocore.xform_name(secondary_attribute.meta.resource_model.name),
                **_clean_boto3_metadata(secondary_attribute.meta.data),
            )

    @property
    def secondary_attribute_models(self) -> Generator[Action, None, None]:
        """Return the secondary attribute models it's possible for this resource type to have."""
        for subresource_mapping, subresource_model in self._boto3_subresource_models:
            if subresource_mapping.type == "secondaryAttribute":
                yield subresource_model

    @property
    def _boto3_subresource_models(self) -> Generator[Tuple[ResourceMap, Action], None, None]:
        """Return the Boto3 subresource mappings and models that exist for this resource type.

        This is used exclusively for secondary attributes because secondary attributes have a 1:1 relationship with
        their parent resource.
        """
        subresource_models: List[Action] = self.boto3_resource.meta.resource_model.subresources  # type: ignore
        for subresource_model in subresource_models:
            subresource_mapping = self.service_map.get_resource_map(subresource_model.name)
            yield subresource_mapping, subresource_model

    @property
    def subresource_summary(self) -> List["ResourceSummary"]:
        """Return a summary of subresource types in this resource."""
        subresources = []
        for subresource_type in self.resource_map.subresource_types:
            resource = self.cloudwanderer_boto3_service._get_empty_resource(subresource_type)
            if not resource:
                continue
            subresources.append(
                ResourceSummary(
                    resource_type=subresource_type,
                    resource_type_pascal=resource.resource_type_pascal,
                    service_friendly_name=self.cloudwanderer_boto3_service.friendly_name,
                    subresources=resource.subresource_summary,
                    secondary_attribute_names=self.secondary_attribute_names,
                )
            )
        return subresources

    def get_subresources(self) -> Generator["CloudWandererBoto3Resource", None, None]:
        """Return the CloudWanderer style subresources of this resource."""
        for subresource_model in self.resource_map.subresource_models:
            collection = getattr(self.boto3_resource, subresource_model.name)
            if not subresource_model.resource:
                logger.error("Could not find resource in %s model", subresource_model.name)
                continue
            resource_map = self.service_map.get_resource_map(subresource_model.resource.model.name)
            for boto3_resource in collection.all():
                if hasattr(boto3_resource, "load") and (
                    not boto3_resource.meta.data or resource_map.requires_load_for_full_metadata
                ):
                    boto3_resource.load()
                yield CloudWandererBoto3Resource(
                    account_id=self.account_id,
                    cloudwanderer_boto3_service=self.cloudwanderer_boto3_service,
                    boto3_resource=boto3_resource,
                    service_map=self.service_map,
                    resource_map=resource_map,
                )

    @property
    def parent_resource_type(self) -> str:
        """Return the resource type of the parent (if it has one)."""
        return self.resource_map.parent_resource_type

    @property
    def is_subresource(self) -> bool:
        """Return whether or not this resource is a subresource."""
        if self.resource_map.type == "baseResource":
            return False
        return bool(len(self.boto3_resource.meta.resource_model.identifiers) == 2)

    @memoized_method()
    def _get_region(self) -> str:
        """Return the region for a resource which requires an API call to determine its region.

        Raises:
            ValueError: When an expected region request definition is missing
        """
        region_request_definition = self.resource_map.region_request
        if region_request_definition is None:
            raise ValueError(
                f"{self.resource_type} is expected to have a region request definition in its service map."
            )
        method = getattr(self._boto3_client, region_request_definition.operation)
        result = method(**region_request_definition.build_params(self.boto3_resource))
        return (
            jmespath.search(region_request_definition.path_to_region, result) or region_request_definition.default_value
        )

    @property
    def _boto3_client(self) -> botocore.client.BaseClient:
        return self.boto3_resource.meta.client


class ResourceSummary(NamedTuple):
    """A summary of a resource's subresource types and secondary attribute names."""

    resource_type: str
    resource_type_pascal: str
    service_friendly_name: str
    subresources: List["ResourceSummary"]  # type: ignore
    secondary_attribute_names: List[str]



"""Helpers that facilitate type hints and mypy validation."""
import functools
import weakref
from threading import RLock
from typing import Any, Callable, Optional


def memoized_method(*lru_args, **lru_kwargs) -> Callable:
    """Instance oriented wrapper of lru_cache.

    from https://stackoverflow.com/questions/33672412/python-functools-lru-cache-with-class-methods-release-object

    Arguments:
        *lru_args:
            The args that will be passed on to lru_cache
        **lru_kwargs:
            The kwargs that will be passed on to lru_cache
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapped_func(self: Any, *args, **kwargs) -> Callable:
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs) -> Any:
                return func(self_weak(), *args, **kwargs)

            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)

        return wrapped_func

    return decorator


_NOT_FOUND = object()


class cached_property:
    """Instance oriented caching of a property method."""

    def __init__(self, func: Callable) -> None:
        self.func = func
        self.attrname: Optional[str] = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, owner: Any, name: str) -> None:
        """Determine the name of the attribute we're decorating.

        Arguments:
            owner: The class the method we're decorating is a member of
            name: The name of the attribute we're decorating

        Raises:
            TypeError: When the same cached_property is being applied to two different names.
        """
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names " f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance: Any, owner: Any = None) -> Any:
        """Return the cached value if one exists.

        Arguments:
            instance: The object our decorated method is a member of
            owner: The class our decorated method is a member of

        Raises:
            TypeError: When cached_property is being used without calling __set_name__ first.
        """
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError("Cannot use cached_property instance without calling __set_name__ on it.")
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                            f"does not support item assignment for caching {self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        return val



"""Module containing abstract classes for CloudWanderer storage connectors."""
from abc import ABC, abstractmethod
from typing import Iterator, List, Optional

from ..cloud_wanderer_resource import CloudWandererResource
from ..urn import URN


class BaseStorageConnector(ABC):
    """Abstract class for specification of the CloudWanderer storage connector interface."""

    @abstractmethod
    def init(self) -> None:
        """Initialise the storage backend whatever it is."""

    @abstractmethod
    def write_resource(self, resource: CloudWandererResource) -> None:
        """Persist a single resource to storage.

        Arguments:
            resource (CloudWandererResource): The CloudWandererResource to write.
        """

    @abstractmethod
    def read_all(self) -> Iterator[dict]:
        """Return all records from storage."""

    @abstractmethod
    def read_resource(self, urn: URN) -> Optional[CloudWandererResource]:
        """Return a resource matching the supplied urn from storage.

        Arguments:
            urn (URN): The AWS URN of the resource to return
        """

    @abstractmethod
    def read_resources(
        self,
        account_id: str = None,
        region: str = None,
        service: str = None,
        resource_type: str = None,
        urn: URN = None,
    ) -> Iterator["CloudWandererResource"]:
        """Yield a resource matching the supplied urn from storage.

        All arguments are optional.

        Arguments:
            urn: The AWS URN of the resource to return
            account_id: AWS Account ID
            region: AWS region (e.g. ``'eu-west-2'``)
            service: Service name (e.g. ``'ec2'``)
            resource_type: Resource Type (e.g. ``'instance'``)
        """

    @abstractmethod
    def delete_resource(self, urn: URN) -> None:
        """Delete this resource and all its resource attributes.

        Arguments:
            urn (URN): The URN of the resource to delete
        """

    @abstractmethod
    def delete_resource_of_type_in_account_region(
        self, service: str, resource_type: str, account_id: str, region: str, urns_to_keep: List[URN] = None
    ) -> None:
        """Delete resources of type in account and region unless in list of URNs.

        This is used primarily to clean up old resources.

        Arguments:
            account_id (str): AWS Account ID
            region (str): AWS region (e.g. ``'eu-west-2'``)
            service (str): Service name (e.g. ``'ec2'``)
            resource_type (str): Resource Type (e.g. ``'instance'``)
            urns_to_keep (List[cloudwanderer.urn.URN]): A list of resources not to delete
        """



"""Allows CloudWanderer to store resources in memory."""
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional

from ..cloud_wanderer_resource import CloudWandererResource
from ..urn import URN
from ..utils import standardise_data_types
from .base_connector import BaseStorageConnector

logger = logging.getLogger(__name__)


class MemoryStorageConnector(BaseStorageConnector):
    """Storage connector to place data in memory.

    Useful for testing.

    Example:
        >>> import cloudwanderer
        >>> cloud_wanderer = cloudwanderer.CloudWanderer(
        ...     storage_connectors=[cloudwanderer.storage_connectors.MemoryStorageConnector()]
        ... )

    """

    def __init__(self) -> None:
        self._data: Dict[str, Any] = {}

    def init(self) -> None:
        """Do nothing. Dummy method to fulfil interface requirements."""

    def read_resource(self, urn: URN) -> Optional[CloudWandererResource]:
        try:
            return memory_item_to_resource(urn, self._data[str(urn)], loader=self.read_resource)
        except KeyError:
            return None

    def read_resources(
        self,
        account_id: str = None,
        region: str = None,
        service: str = None,
        resource_type: str = None,
        urn: URN = None,
    ) -> Iterator["CloudWandererResource"]:
        for urn_str, items in self._data.items():
            item_urn = URN.from_string(urn_str)
            if urn is not None:
                if item_urn == urn:
                    yield memory_item_to_resource(item_urn, items, loader=self.read_resource)
                continue
            if self._urn_matches_kwargs(
                item_urn,
                account_id=account_id,
                region=region,
                service=service,
                resource_type=resource_type,
            ):
                yield memory_item_to_resource(item_urn, items, loader=self.read_resource)

    def _urn_matches_kwargs(self, urn: URN, **kwargs) -> bool:
        filter_items = ("account_id", "region", "service", "resource_type")
        for item in filter_items:
            if kwargs.get(item) is not None and getattr(urn, item) != kwargs[item]:
                return False
        return True

    def read_all(self) -> Iterator[dict]:
        """Return the raw dictionaries stored in memory."""
        for urn_str, items in self._data.items():
            for item_type, item in items.items():
                item_dict = item if isinstance(item, dict) else {"value": item}
                yield {
                    **{
                        "urn": urn_str,
                        "attr": item_type,
                    },
                    **item_dict,
                }

    def write_resource(self, resource: CloudWandererResource) -> None:
        self._data[str(resource.urn)] = self._data.get(str(resource.urn), {})
        self._data[str(resource.urn)]["BaseResource"] = standardise_data_types(
            resource.cloudwanderer_metadata.resource_data
        )

        self._data[str(resource.urn)]["ParentUrn"] = resource.parent_urn
        self._data[str(resource.urn)]["SubresourceUrns"] = resource.subresource_urns

        for secondary_attribute in resource.cloudwanderer_metadata.secondary_attributes:
            self._write_secondary_attribute(
                urn=resource.urn, attribute_type=secondary_attribute.name, secondary_attribute=secondary_attribute
            )

    def _write_secondary_attribute(self, urn: URN, attribute_type: str, secondary_attribute: Dict[str, Any]) -> None:
        """Write the specified resource attribute to storage.

        Arguments:
            urn (URN): The resource whose attribute to write.
            attribute_type (str): The type of the resource attribute to write (usually the boto3 client method name)
            secondary_attribute (boto3.resources.base.ServiceResource): The resource attribute to write to storage.

        """
        self._data[str(urn)] = self._data.get(str(urn), {})
        self._data[str(urn)][attribute_type] = secondary_attribute

    def delete_resource(self, urn: URN) -> None:
        try:
            del self._data[str(urn)]
        except KeyError:
            pass
        subresources_to_delete = set()
        for subresource_urn, resource_dict in self._data.items():
            if resource_dict.get("ParentUrn") == urn:
                subresources_to_delete.add(subresource_urn)
        for subresource_urn in subresources_to_delete:
            del self._data[subresource_urn]

    def delete_resource_of_type_in_account_region(
        self, service: str, resource_type: str, account_id: str, region: str, urns_to_keep: List[URN] = None
    ) -> None:
        urns_to_keep = urns_to_keep or []
        urns_to_delete = []
        for urn_str, items in self._data.items():
            urn = URN.from_string(urn_str)
            if urn.service != service:
                continue
            if urn.account_id != account_id:
                continue
            if urn.resource_type != resource_type:
                continue
            if urn.region != region:
                continue
            if urn in urns_to_keep:
                continue
            urns_to_delete.append(urn)
        for urn in urns_to_delete:
            del self._data[str(urn)]

    def __repr__(self) -> str:
        """Return an instantiable string representation of this object."""
        return f"{self.__class__.__name__}()"

    def __str__(self) -> str:
        """Return a string representation of this object."""
        return f"<{self.__class__.__name__}>"


def memory_item_to_resource(urn: URN, items: Dict[str, Any] = None, loader: Callable = None) -> CloudWandererResource:
    """Convert a resource and its attributes to a CloudWandererResource.

    Arguments:
        urn (URN): The URN of the resource.
        items (dict): The dictionary of items stored under this URN. (Secondary Attributs, BaseResource)
        loader (Callable): The method which can be used to fulfil the :meth:`CloudWandererResource.load`

    """
    items = items or {}
    attributes = [
        attribute
        for item_type, attribute in items.items()
        if item_type not in ["SubresourceUrns", "BaseResource", "ParentUrn"]
    ]
    base_resource: Dict[str, Any] = next(
        iter(resource for item_type, resource in items.items() if item_type == "BaseResource"), {}
    )
    return CloudWandererResource(
        urn=urn,
        subresource_urns=items.get("SubresourceUrns"),
        resource_data=base_resource,
        secondary_attributes=attributes,
        loader=loader,
    )



"""I wandered lonely through the cloud."""
from . import cloud_wanderer_resource, storage_connectors
from .aws_interface import CloudWandererAWSInterface
from .cloud_wanderer import CloudWanderer
from .cloud_wanderer_resource import CloudWandererResource
from .models import ResourceFilter
from .urn import URN

__all__ = [
    "storage_connectors",
    "cloud_wanderer_resource",
    "CloudWandererResource",
    "CloudWanderer",
    "URN",
    "CloudWandererAWSInterface",
    "ResourceFilter",
]



"""A standardised interface for interacting with AWS.

Provides simpler methods for :class:`~.cloud_wanderer.CloudWanderer` to call.
"""

import logging
from typing import Any, Dict, Iterator, List, Optional

import boto3
import botocore  # type: ignore

from cloudwanderer.utils import snake_to_pascal

from .boto3_helpers import Boto3CommonAttributesMixin
from .boto3_loaders import ServiceMappingLoader
from .boto3_services import Boto3Services, CloudWandererBoto3Service, MergedServiceLoader
from .cloud_wanderer_resource import CloudWandererResource
from .exceptions import BadRequestError, ResourceNotFoundError, UnsupportedResourceTypeError
from .models import GetAndCleanUp, ResourceFilter
from .urn import URN

logger = logging.getLogger(__name__)


class CloudWandererAWSInterface(Boto3CommonAttributesMixin):
    """Simplifies lookup of Boto3 services and resources."""

    limit_resources = None

    def __init__(
        self,
        boto3_session: boto3.session.Session = None,
        service_loader: MergedServiceLoader = None,
        service_mapping_loader: ServiceMappingLoader = None,
        resource_filters: List[ResourceFilter] = None,
    ) -> None:
        """Simplifies lookup of Boto3 services and resources.

        Arguments:
            boto3_session (boto3.session.Session):
                A Boto3 session, if not provided the default will be used.
            service_loader:
                An optional loader to allow the injection of additional custom services
            service_mapping_loader:
                An optional loader to allow the injection of additional custom service mappings
            resource_filters:
                An optional list of resource filters to apply when getting those resources.
        """
        self.boto3_session = boto3_session or boto3.session.Session()
        self.boto3_services = Boto3Services(
            boto3_session=boto3_session, service_loader=service_loader, service_mapping_loader=service_mapping_loader
        )
        self.resource_filters = resource_filters or []

    def get_resource(self, urn: URN, include_subresources: bool = True) -> Iterator[CloudWandererResource]:
        """Yield the resource picked out by this URN and optionally its subresources.

        Arguments:
            urn (URN): The urn of the resource to get.
            include_subresources: Whether or not to additionally yield the subresources of the resource.
        """
        try:
            resource = self.boto3_services.get_resource_from_urn(urn=urn)
        except ResourceNotFoundError:
            return None
        except BadRequestError:
            logger.debug(
                f"Got BadRequestError while getting {urn}, as AWS services commonly return 4xx errors other than 404 "
                "for resource non-existence we are interpreting this as the resource does not exist."
            )
            return None
        subresource_urns = []
        if include_subresources:
            for subresource in resource.get_subresources():
                subresource_urns.append(subresource.urn)
                yield CloudWandererResource(
                    urn=subresource.urn,
                    resource_data=subresource.normalised_raw_data,
                    secondary_attributes=list(subresource.get_secondary_attributes()),
                )
        yield CloudWandererResource(
            urn=urn,
            subresource_urns=subresource_urns,
            resource_data=resource.normalised_raw_data,
            secondary_attributes=list(resource.get_secondary_attributes()),
        )

    def get_resources(
        self,
        service_name: str,
        resource_type: str,
        region: str = None,
        **kwargs,
    ) -> Iterator[CloudWandererResource]:
        """Return all resources of resource_type from Boto3.

        Arguments:
            service_name (str): The name of the service to get resource for (e.g. ``'ec2'``)
            resource_type (str): The type of resource to get resources of (e.g. ``'instance'``)
            region (str): The region to get resources of (e.g. ``'eu-west-1'``)
            **kwargs: Additional keyword arguments will be passed down to the Boto3 client.

        Raises:
            botocore.exceptions.ClientError: Occurs if the Boto3 Client Errors.
        """
        logger.info("Getting %s %s resources from %s", service_name, resource_type, region)
        service = self.boto3_services.get_service(service_name=service_name, region_name=region)
        resource_filters = self._get_resource_filters(service_name, resource_type)
        try:
            for resource in service.get_resources(resource_type=resource_type, resource_filters=resource_filters):
                logger.debug("Found %s", resource.urn)
                subresource_urns = []
                for subresource in resource.get_subresources():
                    logger.debug("Found subresource %s", subresource.urn)
                    subresource_urns.append(subresource.urn)
                    yield CloudWandererResource(
                        urn=subresource.urn,
                        resource_data=subresource.normalised_raw_data,
                        secondary_attributes=list(subresource.get_secondary_attributes()),
                    )
                yield CloudWandererResource(
                    urn=resource.urn,
                    subresource_urns=subresource_urns,
                    resource_data=resource.normalised_raw_data,
                    secondary_attributes=list(resource.get_secondary_attributes()),
                )
        except botocore.exceptions.EndpointConnectionError:
            logger.info("%s %s not supported in %s", service_name, resource_type, region)
            return
        except botocore.exceptions.ClientError as ex:
            if ex.response["Error"]["Code"] == "InvalidAction":
                logger.info("%s %s not supported in %s", service_name, resource_type, region)
                return
            raise

    def get_actions(
        self,
        regions: List[str] = None,
        service_names: List[str] = None,
        resource_types: List[str] = None,
        exclude_resources: List[str] = None,
    ) -> List[GetAndCleanUp]:
        """Return the query and cleanup actions to be performed based on the parameters provided.

        All arguments are optional.

        Arguments:
            regions(list):
                The name of the region to get resources from (defaults to session default if not specified)
            service_names (str):
                The names of the services to write resources for (e.g. ``['ec2']``)
            resource_types (list):
                A list of resource types to include (e.g. ``['instance']``)
            exclude_resources (list):
                A list of service:resources to exclude (e.g. ``['ec2:instance']``)

        Raises:
            UnsupportedResourceTypeError: When a resource type is requested that does not exist.
        """
        get_and_cleanup_actions = []
        regions = regions or self.boto3_services.enabled_regions
        exclude_resources = exclude_resources or []
        regions = regions or self.enabled_regions
        service_names = service_names or self.boto3_services.available_services

        for region_name in regions:
            for service_name in service_names:
                service = self.boto3_services.get_empty_service(service_name=service_name, region_name=region_name)
                service_resource_types = self._get_resource_types_for_service(service, resource_types)
                for resource_type in service_resource_types:
                    service_resource = f"{service_name}:{resource_type}"
                    logger.debug("Getting actions for %s %s in %s", service_resource, resource_type, region_name)
                    if service_resource in exclude_resources:
                        logger.debug("Skipping %s as per exclude_resources", service_resource)
                        continue
                    if self.limit_resources and service_resource not in self.limit_resources:
                        logger.debug("Skipping %s as it does not exist in limit_resources", service_resource)
                        continue
                    resource_map = service.service_map.get_resource_map(snake_to_pascal(resource_type))
                    if not resource_map:
                        raise UnsupportedResourceTypeError("No %s resource type found", service_resource)
                        continue
                    actions = resource_map.get_and_cleanup_actions(region_name)
                    if not actions:
                        continue
                    resource_actions = actions.inflate_actions(service.enabled_regions)
                    for subresource_type in resource_map.subresource_types:
                        subresource_map = service.service_map.get_resource_map(snake_to_pascal(subresource_type))
                        actions = subresource_map.get_and_cleanup_actions(region_name)
                        resource_actions += actions.inflate_actions(service.enabled_regions)
                    get_and_cleanup_actions.append(resource_actions)
        return get_and_cleanup_actions

    def _get_resource_types_for_service(
        self, service: CloudWandererBoto3Service, resource_types: Optional[List[str]]
    ) -> List[str]:
        if resource_types:
            logger.debug("Validating if %s are %s resource types", resource_types, service.name)
            return list(set(resource_types) & set(service.resource_types))

        return service.resource_types

    def _get_resource_filters(self, service_name: str, resource_type: str) -> Dict[str, Any]:
        if not self.resource_filters:
            return {}

        for resource_filter in self.resource_filters:
            if resource_filter.service_name == service_name and resource_filter.resource_type == resource_type:
                return resource_filter.filters
        return {}



"""Collection of loose utility functions."""
import json
import logging
import os
from datetime import datetime
from decimal import Decimal
from pathlib import Path
from typing import Any, Callable, Dict, Optional

logger = logging.getLogger(__name__)


def exception_logging_wrapper(method: Callable, return_value: object = None, **kwargs) -> Optional[Any]:
    """Log exceptions raised by method_name.

    Increases visibility of multi-threaded exceptions.

    Arguments:
        method (Callable): A method to wrap, for which will we handle and log any exceptions.
        return_value (object): A value to return instead of the method's result.
        **kwargs: Arguments to pass to the method
    """
    try:
        result = method(**kwargs)
    except Exception as ex:
        logger.exception(ex)
        return return_value
    if return_value is None:
        return result
    return return_value


def json_object_hook(dct: dict) -> dict:
    """Clean out empty strings to avoid DynamoDB ValidationExceptions.

    Arguments:
        dct (dict): The dictionary to clean.
    """
    for key, value in dct.items():
        if value == "":
            dct[key] = None
    return dct


def json_default(item: object) -> Optional[object]:
    """JSON object type converter that handles datetime objects.

    Arguments:
        item: The object JSON is trying to serialise.
    """
    if isinstance(item, datetime):
        return item.isoformat()
    return None


def standardise_data_types(resource: dict) -> Dict[str, Any]:
    """Return a dictionary normalised to datatypes acceptable for DynamoDB.

    Arguments:
        resource (dict): The dictionary we're normalising to DynamoDB acceptable data types.
    """
    result = json.loads(json.dumps(resource, default=json_default), object_hook=json_object_hook, parse_float=Decimal)
    return result


def load_json_definitions(path: str) -> Dict[str, Any]:
    """Return the parsed contents of all JSON files in a given path.

    Arguments:
        path: The path to load JSON files from.
    """
    definition_files = [
        (os.path.abspath(os.path.join(path, file_name)), Path(file_name).stem)
        for file_name in os.listdir(path)
        if os.path.isfile(os.path.join(path, file_name))
    ]
    definitions = {}
    for file_path, service_name in definition_files:
        with open(file_path) as definition_path:
            definitions[service_name] = json.load(definition_path)
    return definitions


def snake_to_pascal(snake_case: str) -> str:
    """Return a PascalCase version of a snake_case name.

    Arguments:
        snake_case: the string to turn into PascalCase.
    """
    snake_case = snake_case.lower()
    return snake_case.replace("_", " ").title().replace(" ", "")
